#define MICROPY_HW_BOARD_NAME "Wio Terminal D51R"
#define MICROPY_HW_MCU_NAME   "SAMD51P19A"

#define MICROPY_HW_XOSC32K          (1)

#define MICROPY_HW_DEFAULT_UART_ID  (2)
#define MICROPY_HW_DEFAULT_I2C_ID   (4)
#define MICROPY_HW_DEFAULT_SPI_ID   (5)

#define MICROPY_HW_QSPIFLASH        W25Q32JV_IQ
