#define MICROPY_HW_BOARD_NAME "Seeed Xiao"
#define MICROPY_HW_MCU_NAME   "SAMD21G18A"

#define MICROPY_HW_XOSC32K  (1)
#define MICROPY_HW_ADC_VREF (2)

#define MICROPY_HW_DEFAULT_UART_ID  (4)
#define MICROPY_HW_DEFAULT_I2C_ID   (2)
#define MICROPY_HW_DEFAULT_SPI_ID   (0)
