#define MICROPY_HW_BOARD_NAME "SAMD21-XPLAINED-PRO"
#define MICROPY_HW_MCU_NAME   "SAMD21J18A"

#define MICROPY_HW_XOSC32K  (1)

#define MICROPY_HW_SPIFLASH (1)
#define MICROPY_HW_SPIFLASH_ID (5)
#define MICROPY_HW_SPIFLASH_BAUDRATE (12000000)
