// Board and hardware specific configuration
#define MICROPY_HW_BOARD_NAME "SparkFun Pro Micro RP2350"
#define MICROPY_HW_FLASH_STORAGE_BYTES (14 * 1024 * 1024)

#define MICROPY_HW_USB_VID (0x1B4F)
#define MICROPY_HW_USB_PID (0x0039)

#define MICROPY_HW_UART0_TX (0)
#define MICROPY_HW_UART0_RX (1)
#define MICROPY_HW_UART0_CTS (2)
#define MICROPY_HW_UART0_RTS (3)

#define MICROPY_HW_UART1_TX (8)
#define MICROPY_HW_UART1_RX (9)
#define MICROPY_HW_UART1_CTS (6)
#define MICROPY_HW_UART1_RTS (7)

#define MICROPY_HW_I2C0_SDA  (16)
#define MICROPY_HW_I2C0_SCL  (17)

#define MICROPY_HW_SPI0_SCK  (22)
#define MICROPY_HW_SPI0_MOSI (23)
#define MICROPY_HW_SPI0_MISO (20)

#define MICROPY_HW_PSRAM_CS_PIN (19)
#define MICROPY_HW_ENABLE_PSRAM (1)
