#define MICROPY_HW_BOARD_NAME "Mini SAM M4"
#define MICROPY_HW_MCU_NAME   "SAMD51G19A"

#define MICROPY_HW_DFLL_USB_SYNC    (1)

#define MICROPY_HW_DEFAULT_UART_ID  (3)
#define MICROPY_HW_DEFAULT_I2C_ID   (2)
#define MICROPY_HW_DEFAULT_SPI_ID   (1)

#define MICROPY_HW_QSPIFLASH   GD25Q16C
