// This file is needed by ospi_xip/source/ospi/ospi_drv.c.
#define OSPI_XIP_ENABLE_AES_DECRYPTION (0)
#define OSPI_XIP_RX_SAMPLE_DELAY (3)
#define OSPI_XIP_DDR_DRIVE_EDGE (1)
#define OSPI_XIP_RXDS_DELAY (12)
